@@ -25,7 +25,7 @@
 // and as implemented in the kernel module of the Digraphs package for GAP in
 // C:
 //
-// https://github.com/gap-packages/Digraphs/blob/master/src/schreier-sims.c
+// https://github.com/digraphs/Digraphs/blob/master/src/schreier-sims.c
 //
 // by Wilf A. Wilson.
 
@@ -57,16 +57,16 @@
 #include <string>         // for operator+, basic_string
 #include <unordered_set>  // for unordered_set
 
-#include "adapters.hpp"                 // for action, degree, inverse
-#include "bruidhinn-traits.hpp"         // for detail::BruidhinnTraits
-#include "containers.hpp"               // for Array2, StaticVector2
-#include "int-range.hpp"                // for IntegralRange
-#include "libsemigroups-config.hpp"     // for LIBSEMIGROUPS_HPCOMBI
-#include "libsemigroups-debug.hpp"      // for LIBSEMIGROUPS_ASSERT
-#include "libsemigroups-exception.hpp"  // for LIBSEMIGROUPS_EXCEPTION
-#include "stl.hpp"                      // for EqualTo
-#include "transf.hpp"                   // for Perm
-#include "types.hpp"                    // for SmallestInteger
+#include "adapters.hpp"          // for action, degree, inverse
+#include "bruidhinn-traits.hpp"  // for detail::BruidhinnTraits
+#include "config.hpp"            // for LIBSEMIGROUPS_HPCOMBI_ENABLED
+#include "containers.hpp"        // for Array2, StaticVector2
+#include "debug.hpp"             // for LIBSEMIGROUPS_ASSERT
+#include "exception.hpp"         // for LIBSEMIGROUPS_EXCEPTION
+#include "int-range.hpp"         // for IntegralRange
+#include "stl.hpp"               // for EqualTo
+#include "transf.hpp"            // for Perm
+#include "types.hpp"             // for SmallestInteger
 
 namespace libsemigroups {
 
@@ -82,21 +82,22 @@ namespace libsemigroups {
   //! permutation group to be represented by a SchreierSims instance.
   //! \tparam TPointType the type of the points acted on.
   //! \tparam TElementType the type of the group elements acting on
-  //! ``TPointType``
+  //! \c TPointType.
   template <size_t N, typename TPointType, typename TElementType>
   struct SchreierSimsTraits {
     //! The type of indices to be used by a SchreierSims instance.
     using index_type = size_t;
 
-    //! The type of the object containing all points acted on by a SchreierSims
-    //! instance.
+    //! Type of the object containing all points acted on.
     using domain_type = IntegralRange<TPointType, 0, N>;
 
+    //! Type of the points acted on.
+    //!
     //! The type of the points acted on by the group represented by \c this,
-    //! which is the same as the template parameter ``TPointType``.
+    //! which is the same as the template parameter \c TPointType.
     using point_type = TPointType;
 
-    //! The type of the group elements acting on ``TPointType``
+    //! Type of the elements.
     using element_type = TElementType;
 
     //! \copydoc libsemigroups::ImageRightAction
@@ -129,11 +130,12 @@ namespace libsemigroups {
   //! \tparam N the largest point not fixed by the permutations in the
   //! permutation group to be represented by this.
   //! \tparam TPointType the type of the points acted on (default:
-  //! SmallestInteger<N>::type).
+  //! the member \c type of SmallestInteger with template parameter \p N).
   //! \tparam TElementType the type of the group elements acting on
-  //! ``TPointType`` (default: NewPermHelper<N>::type)
-  //! \tparam TTraits the type of traits object (default: SchreierSimsTraits<N,
-  //! TPointType, TElementType>)
+  //! \c TPointType (default: the member \c type of \ref LeastPerm with template
+  //! parameter \p N).
+  //! \tparam TTraits the type of traits object (default: SchreierSimsTraits
+  //! with template parameters \c N, \c TPointType, and \c TElementType).
   //!
   //! \sa SchreierSimsTraits.
   //!
@@ -147,7 +149,7 @@ namespace libsemigroups {
   //!  \endcode
   template <size_t N,
             typename TPointType   = typename SmallestInteger<N>::type,
-            typename TElementType = typename NewPermHelper<N>::type,
+            typename TElementType = LeastPerm<N>,
             typename TTraits = SchreierSimsTraits<N, TPointType, TElementType>>
   class SchreierSims final : private detail::BruidhinnTraits<TElementType> {
     static_assert(
@@ -167,26 +169,30 @@ namespace libsemigroups {
         TElementType>::internal_const_value_type;
 
    public:
+    //! Type of the elements.
+    //!
     //! The type of the elements of a SchreierSims instance with const removed,
-    //! and if TElementType is a pointer to const, then the second const is
+    //! and if \c TElementType is a pointer to const, then the second const is
     //! also removed.
     using element_type =
         typename detail::BruidhinnTraits<TElementType>::value_type;
 
-    //! The type of the points acted on by the group represented by \c this,
-    //! which is the same as the template parameter TPointType.
+    //! Type of the points acted on.
+    //!
+    //! Also the template parameter \p TPointType.
     using point_type = TPointType;
 
-    //! The type of the object containing all points acted on by a SchreierSims
-    //! instance.
+    //! Type of the object containing all points acted on.
     using domain_type = typename TTraits::domain_type;
 
-    //! The type of indices to be used by a SchreierSims instance.
+    //! Type of indices.
     using index_type = typename TTraits::index_type;
 
-    // gcc apparently requires the extra qualification on the aliases below
-    //! \copydoc libsemigroups::Action
+    //! Alias for \p TTraits::Action.
+    //!
+    //! See Action for further details.
     using Action = typename TTraits::Action;
+
     //! \copydoc libsemigroups::Degree
     using Degree = typename TTraits::Degree;
     //! \copydoc libsemigroups::EqualTo
@@ -210,13 +216,18 @@ namespace libsemigroups {
     };
 
    public:
+    //! Default constructor.
+    //!
     //! Construct a SchreierSims object representing the trivial group.
     //!
     //! \complexity
     //! \f$O(N ^ 2)\f$ where \p N is the first template parameter.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     SchreierSims()
         : _base(),
           _base_size(0),
@@ -252,7 +263,7 @@ namespace libsemigroups {
     //! Deleted.
     SchreierSims& operator=(SchreierSims&&) = delete;
 
-    //! Add a generator to this.
+    //! Add a generator.
     //!
     //! \param x a const reference to the generator to add.
     //!
@@ -267,7 +278,9 @@ namespace libsemigroups {
     void add_generator(const_element_reference x) {
       if (!has_valid_degree(x)) {
         LIBSEMIGROUPS_EXCEPTION(
-            "generator degree incorrect, expected %d, got %d", N, Degree()(x));
+            "generator degree incorrect, expected %llu, got %llu",
+            uint64_t(N),
+            uint64_t(Degree()(x)));
       } else if (!contains(x)) {
         _finished = false;
         _strong_gens.push_back(0, this->internal_copy(_tmp_element2));
@@ -285,7 +298,7 @@ namespace libsemigroups {
     //!
     //! \complexity
     //! Constant.
-    const_element_reference generator(index_type const index) const {
+    const_element_reference generator(index_type index) const {
       return strong_generator(0, index);
     }
 
@@ -300,10 +313,10 @@ namespace libsemigroups {
     //! \complexity
     //! Constant.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
-    size_t nr_generators() const noexcept {
-      return nr_strong_generators(0);
+    size_t number_of_generators() const noexcept {
+      return number_of_strong_generators(0);
     }
 
     //! The number of strong generators at a given depth.
@@ -314,16 +327,22 @@ namespace libsemigroups {
     //! The number of strong generators, a value of \c size_t, at depth \p
     //! depth of the stabiliser chain.
     //!
-    //! \exceptions
-    //! \noexcept
+    //! \throws LibsemigroupsException if the \p depth is out of bounds.
     //!
     //! \complexity
     //! Constant.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
-    // TODO(later) shouldn't this throw if depth is out of bounds??
-    size_t nr_strong_generators(index_type const depth) const noexcept {
+    size_t number_of_strong_generators(index_type depth) const {
+      if (_base_size == 0) {
+        return 0;
+      } else if (depth >= _base_size) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "depth out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(_base_size),
+            uint64_t(depth));
+      }
       return _strong_gens.size(depth);
     }
 
@@ -336,31 +355,146 @@ namespace libsemigroups {
     //! A const reference to the strong generator of \c this at depth \p depth
     //! and with index \p index.
     //!
+    //! \throws LibsemigroupsException if the \p depth is out of bounds.
     //! \throws LibsemigroupsException if the \p index is out of bounds.
     //!
     //! \complexity
     //! Constant.
-    // TODO(later) shouldn't this throw if depth is out of bounds??
-    const_element_reference strong_generator(index_type const depth,
-                                             index_type const index) const {
+    const_element_reference strong_generator(index_type depth,
+                                             index_type index) const {
+      if (depth >= _base_size) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "depth out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(_base_size),
+            uint64_t(depth));
+      }
       if (index >= _strong_gens.size(depth)) {
         LIBSEMIGROUPS_EXCEPTION(
-            "index out of bounds, expected value in range (0, %d], got %d",
-            _strong_gens.size(depth),
-            index);
+            "index out of bounds, expected value in range (0, %llu], got %llu",
+            uint64_t(_strong_gens.size(depth)),
+            uint64_t(index));
       }
       return this->to_external_const(_strong_gens.at(depth, index));
     }
 
+    //! Get a transversal element.
+    //!
+    //! \param depth the depth.
+    //! \param pt the image of the base point under the traversal.
+    //!
+    //! \returns
+    //! A const reference to the transversal element of \c this at depth
+    //! \p depth moving the corresponding basepoint to the point \p pt.
+    //!
+    //! \throws LibsemigroupsException if the \p depth is out of bounds.
+    //! \throws LibsemigroupsException if \p pt is not in the orbit of the
+    //! basepoint.
+    //!
+    //! \complexity
+    //! Constant.
+    const_element_reference transversal_element(index_type depth,
+                                                point_type pt) {
+      if (depth >= _base_size) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "depth out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(_base_size),
+            uint64_t(depth));
+      }
+      if (pt >= N) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "pt out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(N),
+            uint64_t(pt));
+      }
+      if (!_orbits_lookup[depth][pt]) {
+        LIBSEMIGROUPS_EXCEPTION("no element maps %llu to %llu at depth %llu",
+                                uint64_t(_base[depth]),
+                                uint64_t(pt),
+                                uint64_t(depth));
+      }
+      return this->to_external_const(_transversal[depth][pt]);
+    }
+
+    //! Get an inversal element.
+    //!
+    //! \param depth the depth.
+    //! \param pt the point to map to the base point under the inversal.
+    //!
+    //! \returns
+    //! A const reference to the insversal element of \c this at depth
+    //! \p depth moving the corresponding point \p pt to the basepoint.
+    //!
+    //! \throws LibsemigroupsException if the \p depth is out of bounds.
+    //! \throws LibsemigroupsException if \p pt is not in the orbit of the
+    //! basepoint.
+    //!
+    //! \complexity
+    //! Constant.
+    const_element_reference inversal_element(index_type depth, point_type pt) {
+      if (depth >= _base_size) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "depth out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(_base_size),
+            uint64_t(depth));
+      }
+      if (pt >= N) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "pt out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(N),
+            uint64_t(pt));
+      }
+      if (!_orbits_lookup[depth][pt]) {
+        LIBSEMIGROUPS_EXCEPTION("no element maps %llu to %llu at depth %llu",
+                                uint64_t(_base[depth]),
+                                uint64_t(pt),
+                                uint64_t(depth));
+      }
+      return this->to_external_const(_inversal[depth][pt]);
+    }
+
+    //! Check if a point is in the orbit of a basepoint.
+    //!
+    //! \param depth the depth.
+    //! \param pt the point.
+    //!
+    //! \returns
+    //! A boolean indicating if the point \p pt is in the orbit of the
+    //! basepoint of \c this at depth \p depth.
+    //!
+    //! \throws LibsemigroupsException if the \p depth is out of bounds or if
+    //! \p pt is out of bounds.
+    //!
+    //! \complexity
+    //! Constant.
+    bool orbits_lookup(index_type depth, point_type pt) {
+      if (depth >= _base_size) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "depth out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(_base_size),
+            uint64_t(depth));
+      }
+      if (pt >= N) {
+        LIBSEMIGROUPS_EXCEPTION(
+            "pt out of bounds, expected value in range [0, %llu), got %llu",
+            uint64_t(N),
+            uint64_t(pt));
+      }
+      return _orbits_lookup[depth][pt];
+    }
+
     //! Check if any generators have been added so far.
     //!
-    //! \returns \c true if `nr_generators() == 0` and \c false otherwise.
+    //! \returns \c true if `number_of_generators() == 0` and \c false
+    //! otherwise.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
     //!
     //! \complexity
     //! Constant.
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     // Not noexcept because StaticVector2::size isn't
     bool empty() {
       return _strong_gens.size(0) == 0;
@@ -370,13 +504,16 @@ namespace libsemigroups {
     //!
     //! \returns the size, a value of \c uint64_t.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     uint64_t size() {
       if (empty()) {
         return 1;
       }
-      schreier_sims();
+      run();
       uint64_t out = 1;
       for (index_type i = 0; i < _base_size; i++) {
         out *= _orbits.size(i);
@@ -384,19 +521,20 @@ namespace libsemigroups {
       return out;
     }
 
-    //! Returns a copy of \p x that has been sifted through the stabiliser
-    //! chain.
+    //! Sift an element through the stabiliser chain.
     //!
     //! \param x a const reference to a group element.
     //!
-    //! \returns A value of type SchreierSims::element_type.
+    //! \returns A value of type \ref element_type.
     //!
     //! \throws LibsemigroupsException if the degree of \p x is not equal to
     //! the first template parameter \c N.
     element_type sift(const_element_reference x) {
       if (!has_valid_degree(x)) {
         LIBSEMIGROUPS_EXCEPTION(
-            "element degree incorrect, expected %d, got %d", N, Degree()(x));
+            "element degree incorrect, expected %llu, got %llu",
+            uint64_t(N),
+            uint64_t(Degree()(x)));
       }
       element_type cpy = this->external_copy(x);
       Swap()(cpy, this->to_external(_tmp_element2));
@@ -405,11 +543,14 @@ namespace libsemigroups {
       return cpy;
     }
 
-    //! Returns \c true if \p x belongs to the group.
+    //! Test membership of an element.
     //!
-    //! \param x a const reference to a group element.
+    //! \param x a const reference to the possible element.
     //!
-    //! \returns A bool.
+    //! \returns A \c bool.
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     //!
     //! \note
     //! Returns \c false if the degree of \p x is not equal to the first
@@ -418,7 +559,7 @@ namespace libsemigroups {
       if (!has_valid_degree(x)) {
         return false;
       }
-      schreier_sims();
+      run();
       element_type cpy = this->external_copy(x);
       Swap()(cpy, this->to_external(_tmp_element2));
       this->external_free(cpy);
@@ -426,13 +567,15 @@ namespace libsemigroups {
       return InternalEqualTo()(_tmp_element2, _one);
     }
 
-    //! Returns a const reference to the identity element of the group
-    //! represented by \c this.
+    //! Returns a const reference to the identity.
     //!
-    //! \returns A bool.
+    //! \returns A \c bool.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     const_element_reference identity() const {
       return this->to_external_const(_one);
     }
@@ -448,8 +591,11 @@ namespace libsemigroups {
     //! \complexity
     //! \f$O(N ^ 2)\f$ where \p N is the first template parameter.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     void clear() {
       for (size_t depth = 0; depth < N; ++depth) {
         for (size_t index = 0; index < N; ++index) {
@@ -482,8 +628,11 @@ namespace libsemigroups {
     //! \complexity
     //! Constant.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
     bool finished() const {
       return _finished;
     }
@@ -501,19 +650,20 @@ namespace libsemigroups {
     //!
     //! \complexity
     //! Linear in the current number of base points.
-    void add_base_point(point_type const pt) {
+    void add_base_point(point_type pt) {
       if (pt >= N) {
         LIBSEMIGROUPS_EXCEPTION(
-            "base point out of range, expected value in [0, %d), got %d",
-            N,
-            size_t(pt));
+            "base point out of range, expected value in [0, %llu), got %llu",
+            uint64_t(N),
+            uint64_t(pt));
       } else if (finished()) {
         LIBSEMIGROUPS_EXCEPTION("cannot add further base points");
       } else {
         for (size_t i = 0; i < _base_size; ++i) {
           if (_base[i] == pt) {
             LIBSEMIGROUPS_EXCEPTION(
-                "duplicate base point, %d is already a base point", size_t(pt));
+                "duplicate base point, %llu is already a base point",
+                uint64_t(pt));
           }
         }
       }
@@ -530,12 +680,12 @@ namespace libsemigroups {
     //!
     //! \complexity
     //! Constant.
-    point_type base(index_type const index) const {
+    point_type base(index_type index) const {
       if (index >= _base_size) {
         LIBSEMIGROUPS_EXCEPTION(
-            "index out of bounds, expected value in range (0, %d], got %d",
-            _base_size,
-            index);
+            "index out of bounds, expected value in range (0, %llu], got %llu",
+            uint64_t(_base_size),
+            uint64_t(index));
       }
       return _base[index];
     }
@@ -551,94 +701,30 @@ namespace libsemigroups {
     //! \complexity
     //! Constant.
     //!
-    //! \par Parameters
+    //! \parameters
     //! (None)
     size_t base_size() const noexcept {
       return _base_size;
     }
 
-   private:
-    void init() {
-      _base_size = 0;
-      _finished  = false;
-      _orbits_lookup.fill(false);
-    }
-
-    bool has_valid_degree(const_element_reference x) const {
-      return
-#ifdef LIBSEMIGROUPS_HPCOMBI
-          std::is_same<HPCombi::Perm16, element_type>::value ||
-#endif
-          Degree()(x) == N;
-    }
-
-    void internal_add_base_point(point_type const pt) {
-      LIBSEMIGROUPS_ASSERT(_base_size < N);
-      _base[_base_size] = pt;
-      _orbits.push_back(_base_size, pt);
-      _orbits_lookup[_base_size][pt] = true;
-      _transversal[_base_size][pt]   = this->internal_copy(_one);
-      _inversal[_base_size][pt]      = this->internal_copy(_one);
-      _base_size++;
-    }
-
-    void orbit_enumerate(index_type const depth, index_type first = 0) {
-      LIBSEMIGROUPS_ASSERT(depth < _base_size);
-      for (index_type i = first; i < _orbits.size(depth); i++) {
-        for (auto it = _strong_gens.cbegin(depth);
-             it < _strong_gens.cend(depth);
-             ++it) {
-          orbit_add_point(depth, *it, _orbits.at(depth, i));
-        }
-      }
-    }
-
-    void orbit_add_gen(index_type const depth, internal_element_type gen) {
-      LIBSEMIGROUPS_ASSERT(depth < _base_size);
-      // Apply the new generator to existing points in orbits[depth].
-      index_type old_size_orbit = _orbits.size(depth);
-      for (index_type i = 0; i < old_size_orbit; i++) {
-        orbit_add_point(depth, gen, _orbits.at(depth, i));
-      }
-      orbit_enumerate(depth, old_size_orbit);
-    }
-
-    void orbit_add_point(index_type const      depth,
-                         internal_element_type x,
-                         point_type const      pt) {
-      point_type img = Action()(pt, this->to_external_const(x));
-      if (!_orbits_lookup[depth][img]) {
-        _orbits.push_back(depth, img);
-        _orbits_lookup[depth][img] = true;
-        _transversal[depth][img]   = this->internal_copy(_one);
-        Product()(this->to_external(_transversal[depth][img]),
-                  this->to_external_const(_transversal[depth][pt]),
-                  this->to_external_const(x));
-        _inversal[depth][img] = this->to_internal(
-            Inverse()(this->to_external_const(_transversal[depth][img])));
-      }
-    }
-
-    // Changes _tmp_element2 in-place, and returns the depth reached in the
-    // sifting.
-    index_type internal_sift() {
-      for (index_type depth = 0; depth < _base_size; ++depth) {
-        point_type beta
-            = Action()(_base[depth], this->to_external_const(_tmp_element2));
-        if (!_orbits_lookup[depth][beta]) {
-          return depth;
-        }
-        Product()(this->to_external(_tmp_element1),
-                  this->to_external_const(_tmp_element2),
-                  this->to_external_const(_inversal[depth][beta]));
-        Swap()(this->to_external(_tmp_element2),
-               this->to_external(_tmp_element1));
-      }
-      return _base_size;
-    }
-
-    void schreier_sims() {
-      if (_finished) {
+    //! Run the Schreier-Sims algorithm.
+    //!
+    //! \exceptions
+    //! \no_libsemigroups_except
+    //!
+    //! \complexity
+    //! \f$O(N^2\log^3|G|+|T|N^2\log|G|)\f$ time and
+    //! \f$O(N^2\log|G|+|T|N)\f$ space,
+    //! where \p N is the first template parameter, \f$|G|\f$ is the size of
+    //! the group and \f$|T|\f$ is the number of generators of the group.
+    //!
+    //! \parameters
+    //! (None)
+    //!
+    //! \returns
+    //! (None)
+    void run() {
+      if (_finished || _strong_gens.size(0) == 0) {
         return;
       }
 
@@ -655,7 +741,8 @@ namespace libsemigroups {
         }
       }
 
-      index_type first = 0;
+      index_type first = _strong_gens.size(0) - 1;
+      LIBSEMIGROUPS_ASSERT(first < N);
       for (index_type i = 1; i < _base_size + 1; i++) {
         point_type beta      = _base[i - 1];
         index_type old_first = _strong_gens.size(i);
@@ -720,6 +807,86 @@ namespace libsemigroups {
       _finished = true;
     }
 
+   private:
+    void init() {
+      _base_size = 0;
+      _finished  = false;
+      _orbits_lookup.fill(false);
+    }
+
+    bool has_valid_degree(const_element_reference x) const {
+      return
+#ifdef LIBSEMIGROUPS_HPCOMBI_ENABLED
+          std::is_same<HPCombi::Perm16, element_type>::value ||
+#endif
+          Degree()(x) == N;
+    }
+
+    void internal_add_base_point(point_type pt) {
+      LIBSEMIGROUPS_ASSERT(_base_size < N);
+      _base[_base_size] = pt;
+      _orbits.push_back(_base_size, pt);
+      _orbits_lookup[_base_size][pt] = true;
+      _transversal[_base_size][pt]   = this->internal_copy(_one);
+      _inversal[_base_size][pt]      = this->internal_copy(_one);
+      _base_size++;
+    }
+
+    void orbit_enumerate(index_type depth, index_type first = 0) {
+      LIBSEMIGROUPS_ASSERT(depth < _base_size);
+      for (index_type i = first; i < _orbits.size(depth); i++) {
+        for (auto it = _strong_gens.cbegin(depth);
+             it < _strong_gens.cend(depth);
+             ++it) {
+          orbit_add_point(depth, *it, _orbits.at(depth, i));
+        }
+      }
+    }
+
+    void orbit_add_gen(index_type depth, internal_element_type gen) {
+      LIBSEMIGROUPS_ASSERT(depth < _base_size);
+      // Apply the new generator to existing points in orbits[depth].
+      index_type old_size_orbit = _orbits.size(depth);
+      for (index_type i = 0; i < old_size_orbit; i++) {
+        orbit_add_point(depth, gen, _orbits.at(depth, i));
+      }
+      orbit_enumerate(depth, old_size_orbit);
+    }
+
+    void orbit_add_point(index_type            depth,
+                         internal_element_type x,
+                         point_type            pt) {
+      point_type img = Action()(pt, this->to_external_const(x));
+      if (!_orbits_lookup[depth][img]) {
+        _orbits.push_back(depth, img);
+        _orbits_lookup[depth][img] = true;
+        _transversal[depth][img]   = this->internal_copy(_one);
+        Product()(this->to_external(_transversal[depth][img]),
+                  this->to_external_const(_transversal[depth][pt]),
+                  this->to_external_const(x));
+        _inversal[depth][img] = this->to_internal(
+            Inverse()(this->to_external_const(_transversal[depth][img])));
+      }
+    }
+
+    // Changes _tmp_element2 in-place, and returns the depth reached in the
+    // sifting.
+    index_type internal_sift() {
+      for (index_type depth = 0; depth < _base_size; ++depth) {
+        point_type beta
+            = Action()(_base[depth], this->to_external_const(_tmp_element2));
+        if (!_orbits_lookup[depth][beta]) {
+          return depth;
+        }
+        Product()(this->to_external(_tmp_element1),
+                  this->to_external_const(_tmp_element2),
+                  this->to_external_const(_inversal[depth][beta]));
+        Swap()(this->to_external(_tmp_element2),
+               this->to_external(_tmp_element1));
+      }
+      return _base_size;
+    }
+
     typename domain_type::const_iterator
     first_non_fixed_point(internal_const_element_type x) {
       for (auto it = _domain.cbegin(); it < _domain.cend(); ++it) {
